from .data_loading_utils import *
from .single_formula_comparison_utils import *
from .multi_students_multi_problems_scoring_utils import *
from .singleProblemFormulasTypes import Formula
from .single_formula_comparison_utils import is_almost_equivalent,comparing_rel, whether_rel_latex_correct, whether_rel_latex_correct_with_units


from .singleProblemFormulasTypes import ProblemFormulas,Formula
from .singleStudentAnswerTypes import Student_AnswersAndScores_for_SingleProblem
from .single_formula_comparison_utils import whether_rel_latex_correct_with_units_with_only_one_dict_parameter
import multiprocessing
def compare_problemFormula_with_studentAnswer(dct):
    # The keys in dct should be:
    # problemID, studentID, formulaToken, problemFormula, studentAnswerLatex
    parsing_dct = dict(
        rel_latex = dct['problemFormula'].answer_latex,
        answer_latex = dct['studentAnswerLatex'],
        constants_latex_expression = {**dct['problemFormula'].utils_dct['constants'],
                                    **dct['problemFormula'].utils_dct['universe_constants'],
                                    **dct['problemFormula'].utils_dct['units']},
        strict_comparing_inequalities = dct['problemFormula'].utils_dct['strict_comparing_inequalities'],
        epsilon_for_equal = dct['problemFormula'].utils_dct['epsilon_for_equal'],
        # tolerable_diff_fraction = dct['problemFormula'].utils_dct['tolerable_diff_fraction'], deprecated.
        # tolerable_diff_max = dct['problemFormula'].utils_dct['tolerable_diff_max'], deprecated.
        unit_pattern = dct['problemFormula'].utils_dct['unit_pattern'],
        whole_unit_pattern = dct['problemFormula'].utils_dct['whole_unit_pattern'],
        units_conversion_dict = dct['problemFormula'].utils_dct['units_conversion_dict']
    )
    # print(f"units_conversion_dict: {parsing_dct['units_conversion_dict']}")
    whether_correct, description = whether_rel_latex_correct_with_units_with_only_one_dict_parameter(parsing_dct)
    return dict(
        studentID = dct['studentID'],
        problemID = dct['problemID'],
        formulaToken = dct['formulaToken'],
        whether_correct = whether_correct,
        obtained_points = dct['problemFormula'].max_points if whether_correct else 0,
        description = description   
    )
    
    

def multiStudent_compare_multiProblem(problemFormulasList:list,studentsAnswersAndScores:dict,N_process:int, return_detailed_score = False):
    #studentsAnswersAndScores: dict, key is studentID, value is dict,
        # whose key is problemID, value is of type Student_AnswersAndScores_for_SingleProblem
    
    
    # parsing_dcts_lst_root = []
    parsing_dcts_lst = []
    
    for studentID,studentAnswersAndScores in studentsAnswersAndScores.items():
        for problemformula_for_oneproblem in problemFormulasList:
            problemID = problemformula_for_oneproblem.problemID
            if problemID not in studentAnswersAndScores:
                continue
            assert problemID in studentAnswersAndScores, f"Problem ID {problemID} not found in student answers and scores for student {studentID}"
            studentAnswersAndScoresForThisProblem = studentAnswersAndScores[problemID]
            for formulaToken, formula in problemformula_for_oneproblem.Formula_Dct.items():
                for studentAnswerLatex in studentAnswersAndScoresForThisProblem.studentLatexLst:
                    parsing_dcts_lst.append(dict(
                        studentID = studentID,
                        problemID = problemID,
                        formulaToken = formulaToken,
                        problemFormula = formula,
                        studentAnswerLatex = studentAnswerLatex
                    ))

    # with multiprocessing.Pool(processes=N_process) as pool:
    #     output_compare_lst = pool.map(generate_answer_compare_lst, parsing_dcts_lst_root,chunksize=5)

    # for lst in output_compare_lst:
    #     del(lst[0])
    #     parsing_dcts_lst.extend(lst)
                
    with multiprocessing.Pool(processes=N_process) as pool:
        output_lst = pool.map(compare_problemFormula_with_studentAnswer, parsing_dcts_lst,chunksize=5)
    
    for parsing_dct in output_lst:
        studentID = parsing_dct['studentID']
        problemID = parsing_dct['problemID']
        formulaToken = parsing_dct['formulaToken']
        point = parsing_dct['obtained_points']
        if point > studentsAnswersAndScores[studentID][problemID].studentScoreDct[formulaToken]:
            studentsAnswersAndScores[studentID][problemID].studentScoreDct[formulaToken]=point

    for problemFormulas in problemFormulasList:
        problemID = problemFormulas.problemID
        for studentID, student_answersAndScores_for_SingleProblem in studentsAnswersAndScores.items():
            if problemID in student_answersAndScores_for_SingleProblem:
                if problemID not in student_answersAndScores_for_SingleProblem:
                    continue
                # assert problemID in student_answersAndScores_for_SingleProblem, f"Problem ID {problemID} not found in student answers and scores for student {studentID}"
                student_answersAndScores_for_SingleProblem[problemID].points = problemFormulas.evaluate(
                    student_answersAndScores_for_SingleProblem[problemID].studentScoreDct
                )
                if return_detailed_score:
                    student_answersAndScores_for_SingleProblem[problemID].detailed_score = problemFormulas.evaluate_copying_node(
                        student_answersAndScores_for_SingleProblem[problemID].studentScoreDct
                    )

    return
    
    
    
    
    
       
        
    


def filter_and_convert(problem: dict):
    """
    Remove subquestions whose solutions do not contain $$ equations
    Also filter out invalid problems
    """
    #check the required fields
    if "id" not in problem:
        return None
    
    if "context" not in problem:
        print(f"Problem {problem['id']} is missing context.")
        return None
    
    if "subquestions" not in problem:
        if problem.get("solution"):
            problem["subquestions"] = [{
                "letter": "a",
                "subproblem": problem["context"],
                "solution": problem["solution"]
            }]
            problem["context"] = "Read the following problem and provide your answer."
        else:
            print(f"Problem {problem['id']} is missing subquestions or solution.")
            return None
    
    #check all solutions in subquestions
    contain_equation = False
    for subquestion in problem["subquestions"]:
        if "solution" not in subquestion or not isinstance(subquestion["solution"], str):
            print(f"Subquestion in problem {problem['id']} is missing solution or solution is not a string.")
            return None
        #check if the solution contains any $$ equations
        if "$$" in subquestion["solution"]:
            contain_equation = True
    if not contain_equation:
        print(f"Problem {problem['id']} does not contain any $$ equations in solutions.")
        return None       
    valid_subquestions = problem["subquestions"]
    
    #check if the problem or subproblem context contains $ or numerical numbers
    if "$" not in problem["context"] and not any(
        "$" in subquestion.get("subproblem", "") for subquestion in valid_subquestions
    ) and not any(
        any(char.isdigit() for char in subquestion.get("solution", "")) for subquestion in valid_subquestions
    ):
        print(f"Problem {problem['id']} does not contain any $ or numerical numbers in context or subquestions.")
        return None
    
    problem["subquestions"] = valid_subquestions
    return problem


from .singleStudentAnswerTypes import Student_AnswersAndScores_for_SingleProblem
from .singleProblemFormulasTypes import ProblemFormulas


def _build_Problem_Dict(config_file_location):
    print(f"Loading problem configuration from {config_file_location}")
    namespace = {}
    with open(config_file_location, 'r') as file:
        content = file.read()
        exec(content, namespace)
    print("Done!")

    # Find the first dictionary in the namespace and return it
    for name, value in namespace.items():
        
        if isinstance(value, dict) and name.startswith('question'):
            return value


def build_problem_formula(config_file_location,problemName:None,problemID:None):
    problem_dct = _build_Problem_Dict(config_file_location)
    if problemID is None:
        problemID = problem_dct['problemID']
    # print(problem_dct)
    # assert False
    problemFormulas = ProblemFormulas(problem_dct,problemName,problemID,config_file_location)
    
    return problemFormulas








def _build_Student_AnswersAndScores_for_SingleProblem_from_lst(problemFormulas:ProblemFormulas,studentsAnswers_dcts_lst):
    built_dct=dict()
    problemID = problemFormulas.problemID
    
    for studentAnswers_dct in studentsAnswers_dcts_lst:
        studentID = studentAnswers_dct['studentID']
        student_latexList = studentAnswers_dct['latexList']
        student_AnswersAndScores_for_SingleProblem = Student_AnswersAndScores_for_SingleProblem(studentID,problemID,student_latexList,problemFormulas)
        built_dct[str(studentID)]=student_AnswersAndScores_for_SingleProblem
    
    return built_dct


def build_Student_AnswersAndScores_for_SingleProblem_from_pth(problemFormulas:ProblemFormulas,studentsAnswers_dcts_lst_location):
    import torch
    studentsAnswers_dcts_lst = torch.load(studentsAnswers_dcts_lst_location)
    return _build_Student_AnswersAndScores_for_SingleProblem_from_lst(problemFormulas,studentsAnswers_dcts_lst)


def convert_constant_lists_to_sets(d):
    if isinstance(d, dict):
        for k, v in d.items():
            if isinstance(v, dict):
                convert_constant_lists_to_sets(v)
            elif 'constant' in k and isinstance(v, list):
                d[k] = set(v)
    elif isinstance(d, list):
        for item in d:
            convert_constant_lists_to_sets(item)
def build_problem_formulas_for_multiProblems(problemDict:dict):
    convert_constant_lists_to_sets(problemDict)
    return_dict = dict()
    return_list = list()
    for problemID, content in problemDict.items():
        
        if 'problemID' not in content:
            content['problemID'] = problemID
        else:
            assert content['problemID'] == problemID, f"Problem ID mismatch: {content['problemID']} != {problemID}"
        if 'problemName' in content:
            problemName = content['problemName']
            # delete content['problemName'].
            del content['problemName']
        else:
            problemName = problemID

        print(content)
        problemFormulas = ProblemFormulas(content, problemName, problemID, None)
        return_dict[problemID] = problemFormulas
        return_list.append(problemFormulas)
    return return_list




def build_Student_AnswersAndScores_for_MultiProblem_from_dict(problemFormulasList:list, studentsAnswers_dict: dict):
    # studentsAnswers dict: key is studentID, value is dict,
        # whose key is problemID, value is a list of latex strings.
    built_dct = dict()
    for problemFormulas in problemFormulasList:
        problemID = problemFormulas.problemID
        for studentID, studentAnswers_for_multiProblem in studentsAnswers_dict.items():
            if problemID not in studentAnswers_for_multiProblem:
                continue
            assert problemID in studentAnswers_for_multiProblem, f"Problem ID {problemID} not found in student answers and scores for student {studentID}"
            student_Answers_latex_list = studentAnswers_for_multiProblem[problemID]
            if str(studentID) not in built_dct:
                built_dct[str(studentID)] = dict()
            if problemID not in built_dct[str(studentID)]:
                built_dct[str(studentID)][problemID] = Student_AnswersAndScores_for_SingleProblem(
                    studentID, problemID, student_Answers_latex_list, problemFormulas
                )

    return built_dct


from argparse import ArgumentParser
from utils import Formula
from utils import build_problem_formula,build_Student_AnswersAndScores_for_SingleProblem_from_pth, build_Student_AnswersAndScores_for_MultiProblem_from_dict
from utils import multiStudent_compare_multiProblem
from utils import build_problem_formulas_for_multiProblems, build_Student_AnswersAndScores_for_MultiProblem_from_dict
import json
import os


def parse_arguements():
    parser = ArgumentParser()
    parser.add_argument('--problem_formulas_location', type=str, default='examples/questions_config.json')
    parser.add_argument('--students_answers_location', type=str, default='examples/students_answers.json')
    parser.add_argument('--N_process',type=int,default=20)
    
    return parser.parse_args()


if __name__ == '__main__':
    args = parse_arguements()
    print("args:")
    print(args)
    
    problemFormulas = json.load(open(args.problem_formulas_location,'r',encoding='utf-8'))
    studentsAnswers = json.load(open(args.students_answers_location,'r',encoding='utf-8'))
    
    problemsList = build_problem_formulas_for_multiProblems(problemFormulas)
    studentsAnswersAndScores = build_Student_AnswersAndScores_for_MultiProblem_from_dict(problemsList,studentsAnswers)

    multiStudent_compare_multiProblem(problemsList, studentsAnswersAndScores, args.N_process, return_detailed_score=True)

    for studentID, student_answersAndScores_for_problems in studentsAnswersAndScores.items():
        # print("studentID: ", studentID)
        for problemID, this_student_answerAndScore_for_singleProblem in student_answersAndScores_for_problems.items():
            # print("problemID: ", problemID)
            # print("points: ", this_student_answerAndScore_for_singleProblem.points)
            # print("studentScoreDct: ", this_student_answerAndScore_for_singleProblem.studentScoreDct)
            # print("points: ", this_student_answerAndScore_for_singleProblem.points)
            print(f"Student {studentID} Problem {problemID} Points: {this_student_answerAndScore_for_singleProblem.points}")
            print(f"Student {studentID} Problem {problemID} Detailed score: {this_student_answerAndScore_for_singleProblem.detailed_score.checkScores()}")
    
    
    
    
    
    


EPSILON_FOR_EQUAL = 1e-5
UNIT_PATTERN = r"\\unit{(.*?)}"
WHOLE_UNIT_PATTERN = r"(\\unit{.*?})"
UNITS_CONVERSION_DICT = {
                        "\\km": "1000*m",
                        "\\ms": "0.001*s",
                        "\\kg": "1000*g",
                    }

# default epsilon_for_equal = EPSILON_FOR_EQUAL, can be modified by specifying this hyperparameter in questions_config.json.
# default unit_pattern = UNIT_PATTERN, can be modified by specifying this hyperparameter in questions_config.json.
# default whole_unit_pattern = WHOLE_UNIT_PATTERN, can be modified by specifying this hyperparameter in questions_config.json.
# default units_conversion_dict = UNITS_CONVERSION_DICT, can be modified by specifying this hyperparameter in questions_config.json.


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
  name = 'gradePhyX',         # How you named your package folder (MyLib)
  packages = setuptools.find_packages(),   # Chose the same as "name"
  version = '0.0.4',      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = '(1) an Engine to compare two Equations supporting Physics Quantities (i.e. constants, units, substitutions, etc.); and (2) an efficient implementation for grading a bunch of students answers to a bunch of Physics problems (whose scores are defined with tree-based structures), with multiprocessing support.',   # Give a short description about your library
  author = 'gradePhyX authors',                   # Type in your name
  author_email = 'sjzworking@outlook.com',      # Type in your E-Mail
  url = 'https://github.com/JingzheShi/PhysicsFormulaComparison',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/JingzheShi/PhysicsFormulaComparison/archive/v_01.tar.gz',    # I explain this later on
  keywords = ['physics', 'sympy', 'formula comparison'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'sympy==1.13',
          'antlr4-python3-runtime==4.11',
          'numpy==2.1.2',
          'stopit==1.1.2',
          'regex',
          'tqdm',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
  ],
)


from .singleProblemFormulasTypes import ProblemFormulas
from .singleProblemFormulasTypes import Node
class Student_AnswersAndScores_for_SingleProblem():
    def __init__(self,studentID,problemID,latexList,problem_formula:ProblemFormulas):
        self.studentId = studentID
        self.problemID = problemID
        self.studentLatexLst = latexList
        self.points = 0
        self.detailed_score = None
        self.studentScoreDct = dict()
        def build_dct(node:Node):
            if node.ChildrenNodeType == "formula":
                for child in node.children_lst:
                    self.studentScoreDct[child.TokenStr] = 0
            else:
                for child_node in node.children_lst:
                    build_dct(child_node)
        build_dct(problem_formula.root_node)
        

